import struct
import ipaddress

from aiostun import constants

class Attribute:
    def __init__(self, attr_type):
        """init"""
        self.attr_type = attr_type
        self.params = {}

    def get_name(self):
        """return class name"""
        if self.attr_type in constants.ATTR_NAMES:
            return constants.ATTR_NAMES[self.attr_type]
        return "%s (Unknown)" % self.attr_type

    def __str__(self):
        """sting representation"""
        ret = [ self.get_name() ]
        for l in self.to_string():
            ret.append( "\t\t* %s" % l )
        return "\n".join(ret)

    def decode(self, value):
        """decode the value of the attribute"""
        self.params["value"] = value

    def encode(self):
        """to bytes"""
        return self.params["value"]

    def to_string(self):
        """human string representation"""
        return [ "%s" % self.params["value"] ]

# https://www.rfc-editor.org/rfc/rfc3489#section-11.2.1
# 0                   1                   2                   3
#     0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    |x x x x x x x x|    Family     |           Port                |
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#    |                             Address                           |
#    +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
class AttributeAddr(Attribute):
    def to_string(self):
        """human string representation"""
        ret = [ "Protocol Family: %s" % self.params["family"] ]
        ret.append( "IP: %s" % self.params["ip"] )
        ret.append( "Port: %s" % self.params["port"] )
        return ret

    def decode(self, value):
        """decode the attribute"""
        # read family protocol, ipv4 (1) or ipv6 (2)
        (family,) = struct.unpack("!B", value[1:2])
        if family > 0x02 and family < 0x01:
            return False

        # decode port and ip
        (port,) = struct.unpack("!H", value[2:4])
        if family == 0x01:
            ip = "%s" % ipaddress.IPv4Address(value[4:])
        if family == 0x02:
            ip = "%s" % ipaddress.IPv6Address(value[4:])

        self.params["family"] = constants.FAMILY_NAMES[family]
        self.params["port"] = port
        self.params["ip"] = ip

class AttributeStr(Attribute):
    def __init__(self, attr_type, attr_value):
        Attribute.__init__(self, attr_type)
        self.params["value"] = attr_value
    def to_string(self):
        r = self.params["value"]
        r = "%s" % r if isinstance(r, str) else "%s" % r.decode()
        return [ r ]
    def encode(self):
        r = self.params["value"]
        return r.encode() if isinstance(r, str) else r
        

class AttrXorMappedAddr(Attribute):
    def __init__(self):
        Attribute.__init__(self, attr_type=constants.ATTR_XOR_MAPPED_ADDRESS)
    def to_string(self):
        """human string representation"""
        ret = [ "Protocol Family: %s" % self.params["family"] ]
        ret.append( "IP: %s" % self.params["ip"] )
        ret.append( "Port: %s" % self.params["port"] )
        return ret

    def decode(self, value, tid):
        """decode the attribute"""
        # read family protocol, ipv4 (1) or ipv6 (2)
        (family,) = struct.unpack("!B", value[1:2])
        if family > 0x02 and family < 0x01:
            return False

        # decode port
        (port_xor,) = struct.unpack("!H", value[2:4])
        port = port_xor ^ (constants.MAGIC_COOKIE >> 16)

        # prepare key for xor
        if family == 0x01:
            key = struct.pack("!L", constants.MAGIC_COOKIE)
        if family == 0x02:
            key = struct.pack("!L", constants.MAGIC_COOKIE)
            key += struct.pack("!12s", tid)

        # decode ip
        host = bytes(a ^ b for a, b in zip(value[4:], key))
        if family == 0x01:
            ip = "%s" % ipaddress.IPv4Address(host)
        if family == 0x02:
            ip = "%s" % ipaddress.IPv6Address(host)

        self.params["family"] = constants.FAMILY_NAMES[family]
        self.params["port"] = port
        self.params["ip"] = ip

# basic address (ip/port) attributes
class AttrMappedAddr(AttributeAddr):
    def __init__(self):
        Attribute.__init__(self, attr_type=constants.ATTR_MAPPED_ADDRESS)
class AttrOtherAddress(AttributeAddr):
    def __init__(self):
        Attribute.__init__(self, attr_type=constants.ATTR_OTHER_ADDRESS)
class AttrResponseOrigin(AttributeAddr):
    def __init__(self):
        Attribute.__init__(self, attr_type=constants.ATTR_RESPONSE_ORIGIN)
class AttrSourceAddress(AttributeAddr):
    def __init__(self):
        Attribute.__init__(self, attr_type=constants.ATTR_SOURCE_ADDRESS)
class AttrChangedAddress(AttributeAddr):
    def __init__(self):
        Attribute.__init__(self, attr_type=constants.ATTR_CHANGED_ADDRESS)

class AttrChangeRequest(Attribute):
    def __init__(self, changeIp=False, changePort=False):
        Attribute.__init__(self, attr_type=constants.ATTR_CHANGE_REQUEST)
        self.params["change-ip"] = changeIp
        self.params["change-port"] = changePort

    def encode(self):
        flags = 0

        if self.params["change-ip"]: flags += 4
        if self.params["change-port"]: flags += 2

        return struct.pack("!L", flags)

# https://www.rfc-editor.org/rfc/rfc3489#section-11.2.9
# 0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                   0                     |Class|     Number    |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |      Reason Phrase (variable)                                ..
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
class AttrErrorCode(Attribute):
    def __init__(self):
        Attribute.__init__(self, attr_type=constants.ATTR_ERROR_CODE)

    def to_string(self):
        ret = [ "Code: %s" % self.params["code"] ]
        ret.append( "Phrase: %s" % self.params["phrase"])
        return ret

    def decode(self, value):
        err_code = value[2]*100 + value[3]
        err_phrase = value[4:].decode()
        self.params["code"] = err_code
        self.params["phrase"] = err_phrase


# https://www.rfc-editor.org/rfc/rfc3489#section-11.2.8
class AttrIntegrity(AttributeStr):
    def __init__(self, value):
        AttributeStr.__init__(self, attr_type=constants.ATTR_MESSAGE_INTEGRITY, attr_value=value)
    def to_string(self):
         return [ "0x%s" % self.params["value"].hex() ]

class AttrFingerPrint(AttributeStr):
    def __init__(self, value):
        AttributeStr.__init__(self, attr_type=constants.ATTR_FINGERPRINT, attr_value=value)
    def to_string(self):
         return [ "0x%s" % self.params["value"].hex() ]

class AttrNonce(AttributeStr):
    def __init__(self, value):
        AttributeStr.__init__(self, attr_type=constants.ATTR_NONCE, attr_value=value)

class AttrSoftware(AttributeStr):
    def __init__(self, value):
        AttributeStr.__init__(self, attr_type=constants.ATTR_SOFTWARE, attr_value=value)

class AttrRealm(AttributeStr):
    def __init__(self, value):
        AttributeStr.__init__(self, attr_type=constants.ATTR_REALM, attr_value=value)

class AttrUsername(AttributeStr):
    def __init__(self, value):
        AttributeStr.__init__(self, attr_type=constants.ATTR_USERNAME, attr_value=value)





IPPROTO_UDP = 1
IPPROTO_TCP = 2
IPPROTO_TLS = 3

IPPROTO_NAMES = {
    IPPROTO_UDP: "UDP",
    IPPROTO_TCP: "TCP",
    IPPROTO_TLS: "TLS",
}

FAMILY_IP4 = 0x01
FAMILY_IP6 = 0x02

FAMILY_NAMES = {
        FAMILY_IP4: "IPv4",
        FAMILY_IP6: "IPv6"
}

MAGIC_COOKIE = 0x2112A442
STUN_HEADER_SIZE = 20

CLASS_REQUEST = 0
CLASS_INDICATION = 1
CLASS_SUCCESS = 2
CLASS_ERROR = 3

CLASS_NAMES = {
    CLASS_REQUEST: "Request",
    CLASS_INDICATION: "Indication",
    CLASS_SUCCESS: "Success Response",
    CLASS_ERROR: "Error Response",
}

METHOD_BINDING = 1
METHOD_SHARED_SECRET = 2
METHOD_ALLOCATE = 3
METHOD_REFRESH = 4
METHOD_SEND = 6
METHOD_DATA = 7
METHOD_CREATE_PERMISSION = 8
METHOD_CHANNEL_BIND = 9

METHOD_NAMES = {
    METHOD_BINDING: "Binding",
    METHOD_SHARED_SECRET: "SharedSecret",
    METHOD_ALLOCATE: "Allocate",
    METHOD_REFRESH: "Refresh",
}

ATTR_MAPPED_ADDRESS = 0x001
ATTR_RESPONSE_ADDRESS = 0x002
ATTR_CHANGE_REQUEST = 0x003
ATTR_SOURCE_ADDRESS = 0x004 # deprecated
ATTR_CHANGED_ADDRESS = 0x005 # deprecated
ATTR_USERNAME = 0x006
ATTR_PASSWORD = 0x007
ATTR_MESSAGE_INTEGRITY = 0x008
ATTR_ERROR_CODE = 0x009
ATTR_UNKNOWN_ATTRIBUTE = 0x00a
ATTR_REFLECTED_FROM = 0x00b
ATTR_CHANNEL_NUMBER = 0x00c
ATTR_LIFETIME = 0x00d
ATTR_BANDWIDTH = 0x0010
ATTR_XOR_PEER_ADDRESS = 0x0012
ATTR_DATA = 0x0013
ATTR_REALM = 0x0014
ATTR_NONCE = 0x0015
ATTR_XOR_RELAYED_ADDRESS = 0x0016
ATTR_REQUESTED_ADDRESS_FAMILY = 0x0017
ATTR_EVENT_PORT = 0x0018
ATTR_REQUESTED_TRANSPORT = 0x0019
ATTR_DONT_FRAGMENT = 0x001A
ATTR_MESSAGE_INTEGRITY_SHA256  = 0x001C
ATTR_PASSWORD_ALGORITHM = 0x001D
ATTR_USERHASH = 0x001E
ATTR_XOR_MAPPED_ADDRESS = 0x0020
ATTR_RESERVATION_TOKEN = 0x0022
ATTR_PADDING = 0x0026
ATTR_RESPONSE_PORT = 0x0027
ATTR_PASSWORD_ALGORITHMS = 0x8002
ATTR_ALTERNATE_DOMAIN = 0x8003
ATTR_XOR_MAPPED_ADDRESS_OPTIONAL = 0x8020 # optional for the client to use
ATTR_SOFTWARE = 0x8022
ATTR_ALTERNATE_SERVER = 0x8023
ATTR_PRIORITY = 0x0024
ATTR_USE_CANDIDATE = 0x0025
ATTR_CACHE_TIMEOUT = 0x8027
ATTR_FINGERPRINT = 0x8028
ATTR_ICE_CONTROLLED = 0x8029
ATTR_ICE_CONTROLLING = 0x802A
ATTR_RESPONSE_ORIGIN = 0x802B
ATTR_OTHER_ADDRESS = 0x802C

ATTR_NAMES = {
    ATTR_MAPPED_ADDRESS: "MAPPED-ADDRESS",
    ATTR_USERNAME: "USERNAME",
    ATTR_MESSAGE_INTEGRITY: "MESSAGE-INTEGRITY",
    ATTR_ERROR_CODE: "ERROR-CODE",
    ATTR_UNKNOWN_ATTRIBUTE: "UNKNOWN-ATTRIBUTES",
    ATTR_REALM: "REALM",
    ATTR_NONCE: "NONCE",
    ATTR_REQUESTED_ADDRESS_FAMILY: "REQUESTED-ADDRESS-FAMILY",
    ATTR_XOR_MAPPED_ADDRESS: "XOR-MAPPED-ADDRESS",
    ATTR_XOR_MAPPED_ADDRESS_OPTIONAL: "XOR-MAPPED-ADDRESS",
    ATTR_SOFTWARE: "SOFTWARE",
    ATTR_ALTERNATE_SERVER: "ALTERNATE-SERVER",
    ATTR_FINGERPRINT: "FINGERPRINT",
    ATTR_RESPONSE_ORIGIN: "RESPONSE-ORIGIN",
    ATTR_OTHER_ADDRESS: "OTHER-ADDRESS",
    ATTR_SOURCE_ADDRESS: "SOURCE-ADDRESS",
    ATTR_CHANGED_ADDRESS: "CHANGED-ADDRESS"
}



import unittest
import aiostun



class TestDecode(unittest.TestCase):
    def test_valid_classicstun_binding_request(self):
        """decode classic stun Binding Request"""
        codec = aiostun.Codec()

        # Message Type = Binding Success Response
        Binding_Req = "0001"
        # Message Length 
        Binding_Req += "0000"
        # Message Transaction Id
        Binding_Req += "8112a4427a54477269564651786d7749"

        codec.buf = bytes.fromhex(Binding_Req)
        decoded = codec.decode()
       
        self.assertIsNotNone(decoded)

    def test_valid_binding_request(self):
        """decode Binding Request"""
        codec = aiostun.Codec()

        # Message Type = Binding Success Response
        Binding_Req = "0001"
        # Message Length 
        Binding_Req += "0000"
        # Message Cookie
        Binding_Req += "2112a442"
        # Message Transaction Id
        Binding_Req += "7a54477269564651786d7749"

        codec.buf = bytes.fromhex(Binding_Req)
        decoded = codec.decode()
       
        self.assertIsNotNone(decoded)

    def test_valid_binding_success_response(self):
        """decode Binding Success response"""
        codec = aiostun.Codec()

        # Message Type = Binding Success Response
        Binding_Success = "0101"
        # Message Length 
        Binding_Success += "0048"
        # Message Cookie
        Binding_Success += "2112a442"
        # Message Transaction Id
        Binding_Success += "7a54477269564651786d7749"
        # Attribute XOR-MAPPED-ADDRESS
        Binding_Success += "002000080001a8e877ff14ec"
        # Attribute MAPPED-ADDRESS
        Binding_Success += "00010008000189fa56edb0ae"
        # Attribute RESPONSE-ORIGIN
        Binding_Success += "802b000800010050d827fc0f"
        # Attribute SOFTWARE
        Binding_Success += "80220018436f7475726e2d342e352e32202764616e20456964657227"
        # Attribute FINGERPRINT
        Binding_Success += "80280004d7caaa2b"

        codec.buf = bytes.fromhex(Binding_Success)
        decoded = codec.decode()
        
        self.assertIsNotNone(decoded)

    def test_invalid_binding_request(self):
        """decode invalid Binding Request"""
        codec = aiostun.Codec()

        # Message Type = Binding Success Response
        Binding_Req = "0001"
        # Message Length 
        Binding_Req += "0010"
        # Message Cookie
        Binding_Req += "2112a442"
        # Message Transaction Id
        Binding_Req += "7a54477269564651786d7749"

        codec.buf = bytes.fromhex(Binding_Req)
        decoded = codec.decode()
       
        self.assertIsNone(decoded)

    def test_invalid_binding_success_response(self):
        """decode Binding Success response"""
        codec = aiostun.Codec()

        # Message Type = Binding Success Response
        Binding_Success = "0101"
        # Message Length 
        Binding_Success += "0048"
        # Message Cookie
        Binding_Success += "2112a442"
        # Message Transaction Id
        Binding_Success += "7a54477269564651786d7749"
        # Attribute XOR-MAPPED-ADDRESS
        Binding_Success += "002000080001a8e877ff14ec"
        # Attribute MAPPED-ADDRESS
        Binding_Success += "00010008000189fa56edb0ae"
        # Attribute RESPONSE-ORIGIN
        Binding_Success += "802b000800010050d827fc0f"
        # Attribute SOFTWARE, invalid length
        Binding_Success += "80220018436f7475726e2d342e352e"
        # Attribute FINGERPRINT
        Binding_Success += "80280004d7caaa2b"

        codec.buf = bytes.fromhex(Binding_Success)
        decoded = codec.decode()
        
        self.assertIsNone(decoded)


class TestEncode(unittest.TestCase):
    def test_binding_request(self):
        codec = aiostun.Codec()

        # encode message
        req = aiostun.Message(msgclass=aiostun.CLASS_REQUEST, msgmethod=aiostun.METHOD_ALLOCATE, attrs=[])
        msg = codec.encode(req)

        # decode it
        codec.buf = msg
        decoded = codec.decode()
       
        self.assertIsNotNone(decoded)



from aiostun import client
from aiostun import constants
from aiostun import attribute

NETWORK_ERROR = "Network Error"
PROTOCOL_ERROR = "Protocol Error"
OPEN_INTERNET = "Open Internet"
SYMMETRIC_UDP_FIREWALL = "Symmetric Udp Firewall"
FULL_CONE = "Full Cone"
SYMMETRIC_NAT = "Symmetric NAT"
RESTRICTED_NAT = "Restricted NAT"
RESTRICTED_PORT_NAT = "Restricted Port NAT"

class NAT:
    __DEFAULT_STUN_HOST__ = "turn.matrix.org"
    __DEFAULT_STUN_PORT__ = 3478
    async def discover(self, use_classicstun=False, **kwargs):
        """Discovery NAT"""
        if use_classicstun:
            return await self.classic_discover(**kwargs)

        raise Exception("Not yet implemented")

    async def classic_discover(self, stun_host=__DEFAULT_STUN_HOST__, stun_port=__DEFAULT_STUN_PORT__):
        """discover like described in the rfc3489"""
        nat_behavior = {}
        use_classicstun = True

        # Test I: the client sends a STUN Binding Request to a server,
        # without any flags set in the CHANGE-REQUEST attribute,
        # and without the RESPONSE-ADDRESS attribute.
        stun_test = client.Client(host=stun_host, port=stun_port,
                                  family=constants.FAMILY_IP4,
                                  proto=constants.IPPROTO_UDP)
        # connect and get the local ip and port
        await stun_test.connect(remote_addr=False)

        # Send bind request without any flag
        # if no response, the reason can be multiple: UDP blocked ? network issue ? or the server is down ?
        resp_test1 = await stun_test.bind_request(use_classicstun=use_classicstun, remote_addr=(stun_host, stun_port))
        if resp_test1 is None:
           nat_behavior["error"] = NETWORK_ERROR
           return nat_behavior

        # read local port
        (local_ipI, local_portI) = stun_test.get_local_addr()
        nat_behavior["local-ip"] = local_ipI
        nat_behavior["local-port"] = local_portI

        # if the mapped addr  and changed addr is missing, something is wrong
        mappedAddr = resp_test1.get_attribute(attribute.AttrMappedAddr)
        nat_behavior["external-ip"] = mappedAddr.params["ip"]
        nat_behavior["external-port"] = mappedAddr.params["port"]

        changedaddr = resp_test1.get_attribute(attribute.AttrChangedAddress)
        if mappedAddr is None:
            nat_behavior["error"] = PROTOCOL_ERROR
            return nat_behavior
        if changedaddr is None:
            nat_behavior["error"] = PROTOCOL_ERROR
            return nat_behavior

        # Test II: the client sends a Binding Request with both the "change IP" and "change port" flags
        # from the CHANGE-REQUEST attribute set
        attr_changereq = attribute.AttrChangeRequest(changeIp=True, changePort=True)
        resp_test2 = await stun_test.bind_request(use_classicstun=use_classicstun,
                                                  attrs=[attr_changereq],
                                                  remote_addr=(stun_host, stun_port))

        if mappedAddr.params["ip"] == local_ipI and resp_test2 is None:
            nat_behavior["nat"] = SYMMETRIC_UDP_FIREWALL
            return nat_behavior

        if mappedAddr.params["ip"] == local_ipI and resp_test2:
            nat_behavior["nat"] = OPEN_INTERNET
            return nat_behavior

        if mappedAddr.params["ip"] != local_ipI and resp_test2:
            nat_behavior["nat"] = FULL_CONE
            return nat_behavior

        # Test I Again:  but this time, does to the address and port from the CHANGED-ADDRESS attribute
        # from initial binding request
        if mappedAddr.params["ip"] != local_ipI and resp_test2 is None:

            remote_addr = (changedaddr.params["ip"], changedaddr.params["port"])
            resp_test1_again = await stun_test.bind_request(use_classicstun=use_classicstun,
                                                            attrs=[], remote_addr=remote_addr)
            if resp_test1_again is None:
                nat_behavior["error"] = PROTOCOL_ERROR
                return nat_behavior

            mappedAddr_again = resp_test1_again.get_attribute(attribute.AttrMappedAddr)
            if mappedAddr_again is None:
                nat_behavior["error"] = PROTOCOL_ERROR
                return nat_behavior

            if not (mappedAddr_again.params["ip"] == mappedAddr.params["ip"] and mappedAddr_again.params["port"] == mappedAddr.params["port"]):
                nat_behavior["nat"] = SYMMETRIC_NAT
                return nat_behavior

            # Test III: the client sends a Binding Request with only the "change port" flag set.
            attr_changereq3 = attribute.AttrChangeRequest(changeIp=False, changePort=True)
            remote_addr = (changedaddr.params["ip"], changedaddr.params["port"])
            resp_test3 = await stun_test.bind_request(use_classicstun=use_classicstun,
                                                      attrs=[attr_changereq3],
                                                      remote_addr=remote_addr)
            if resp_test3:
                nat_behavior["nat"] = RESTRICTED_NAT
                return nat_behavior
            else:
                nat_behavior["nat"] = RESTRICTED_PORT_NAT
                return nat_behavior



import struct
import asyncio
import string
import random

from aiostun import constants
from aiostun import attribute

def gen_id(length=12):
    """generate random id"""
    chars = string.ascii_lowercase
    chars += string.ascii_uppercase
    return b''.join([random.choice(chars).encode() for i in range(length)])

class Message(object):
    def __init__(self, msgclass, msgmethod, attrs):
        """init"""
        self.msglength = 0
        self.msgmethod = msgmethod
        self.msgclass = msgclass
        self.magic_cookie = constants.MAGIC_COOKIE
        self.transaction_id = gen_id()
        self.attributes = attrs

    def get_class(self):
        """return class name"""
        if self.msgclass in constants.CLASS_NAMES:
            return constants.CLASS_NAMES[self.msgclass]
        return "%s (Unsupported)" % self.msgclass

    def get_method(self):
        """return method name"""
        if self.msgmethod in constants.METHOD_NAMES:
            return constants.METHOD_NAMES[self.msgmethod]
        return "%s (Unsupported)" % self.msgmethod

    def get_attribute(self, atype):
        """get attribute"""
        for attr in self.attributes:
            if isinstance(attr, atype):
                return attr
        return None

    def decode_attrs(self, attrs):
        """decode all attributes"""
        for attr in attrs:
            # decode the value 
            if attr["type"] in [ constants.ATTR_XOR_MAPPED_ADDRESS, constants.ATTR_XOR_MAPPED_ADDRESS_OPTIONAL ]:
                attr_obj = attribute.AttrXorMappedAddr()
                attr_obj.decode(value=attr["value"], tid=self.transaction_id)
                
            elif attr["type"] in [ constants.ATTR_MAPPED_ADDRESS ]:
                attr_obj = attribute.AttrMappedAddr()
                attr_obj.decode(value=attr["value"])

            elif attr["type"] in [ constants.ATTR_OTHER_ADDRESS ]:
                attr_obj = attribute.AttrOtherAddress()
                attr_obj.decode(value=attr["value"])

            elif attr["type"] in [ constants.ATTR_RESPONSE_ORIGIN ]:
                attr_obj = attribute.AttrResponseOrigin()
                attr_obj.decode(value=attr["value"])

            elif attr["type"] in [ constants.ATTR_SOURCE_ADDRESS ]:
                attr_obj = attribute.AttrSourceAddress()
                attr_obj.decode(value=attr["value"])

            elif attr["type"] in [ constants.ATTR_CHANGED_ADDRESS ]:
                attr_obj = attribute.AttrChangedAddress()
                attr_obj.decode(value=attr["value"])

            elif attr["type"] in [ constants.ATTR_SOFTWARE ]:
                attr_obj = attribute.AttrSoftware(attr["value"])

            elif attr["type"] in [ constants.ATTR_FINGERPRINT ]:
                attr_obj = attribute.AttrFingerPrint(attr["value"])

            elif attr["type"] in [ constants.ATTR_ERROR_CODE ]:
                attr_obj = attribute.AttrErrorCode()
                attr_obj.decode(value=attr["value"])

            elif attr["type"] in [ constants.ATTR_NONCE ]:
                attr_obj = attribute.AttrNonce(attr["value"])

            elif attr["type"] in [ constants.ATTR_REALM ]:
                attr_obj = attribute.AttrRealm(attr["value"])

            else:
                print(attr["type"], attr["value"])
                attr_obj = attribute.Attribute(attr["type"])
                attr_obj.decode(value=attr["value"])

            # append to the list
            self.attributes.append(attr_obj)

    def __str__(self):
        """to string representation"""
        if self.magic_cookie>0:
            ret = ["STUN:"]
        else:
            ret = ["CLASSIC-STUN:"]
        ret.append("\tHeader:")
        ret.append("\t\tMessage Type:")
        ret.append("\t\t\tClass: %s" % self.get_class())
        ret.append("\t\t\tMethod: %s" % self.get_method())
        ret.append("\t\tMessage Length: %s" % self.msglength)
        ret.append("\t\tMessage TransactionID: %s" % self.transaction_id.decode())

        if len(self.attributes):
            ret.append("\tAttributes:")
            for attr in self.attributes:
                ret.append("\t\t%s" % attr)
            #ret.append("")

        return "\n".join(ret)

class ClassicMessage(Message):
    def __init__(self, msgclass, msgmethod, attrs):
        Message.__init__(self, msgclass, msgmethod, attrs)
        self.magic_cookie = 0
        self.transaction_id = gen_id(length=16)

class Codec:
    def __init__(self):
        """init"""
        self.buf = b""
        self._queue = asyncio.Queue(0)

    def feed_data(self, data):
        """append data to the buffer"""
        self.buf = b''.join([self.buf, data])

        resp = self.decode()
        if resp is None: return

        self._queue.put_nowait(resp)

    def decode(self):
        """decode data from buffer"""
        if len(self.buf) < constants.STUN_HEADER_SIZE:
            return None

        # enough data to decode header
        (stunlength,) = struct.unpack("!H", self.buf[2:4])

        if len(self.buf) < stunlength + constants.STUN_HEADER_SIZE:
            return None

        # decode header
        (stuntype, stunlength) = struct.unpack("!HH", self.buf[:4])

        # remote packet from buffer
        pl = self.buf[:stunlength+constants.STUN_HEADER_SIZE]
        self.buf = self.buf[stunlength+constants.STUN_HEADER_SIZE:]

        # decode class and method
        stunclass = ((stuntype & 0x0010) >> 4) | ((stuntype & 0x0100) >> 7)
        stunmethod = (stuntype & 0x000F) | ((stuntype & 0x00E0) >> 1)  | ((stuntype & 0x3E00) >> 2)

        # read magic cookie and transactionid
        (magic_cookie,) = struct.unpack("!L", pl[4:8])
        if magic_cookie != constants.MAGIC_COOKIE:
            magic_cookie = 0
            (transaction_id,) = struct.unpack("!16s", pl[4:20])
        else:
            (transaction_id,) = struct.unpack("!12s", pl[8:20])

        # finally, decode attributes
        pl = pl[20:]
        attrs = []
        while len(pl) >= 4:
            # read attribute
            (attr_type, attr_length,) = struct.unpack("!HH", pl[:4])

            # padding ? always a multiple of 4 bytes
            pad_mod = 4-((attr_length+4) % 4)
            pad_length = 0 if pad_mod == 4 else pad_mod

            attrs.append( {"type": attr_type, "value": pl[4:4+attr_length]} )

            # data remaining for next attributes
            pl = pl[4+attr_length+pad_length:]

        rsp = Message(stunclass, stunmethod, [])
        rsp.msglength = stunlength
        rsp.magic_cookie = magic_cookie
        rsp.transaction_id = transaction_id
        rsp.decode_attrs(attrs)

        return rsp

    def encode(self, m):
        """encode the stun message"""
        # encode attributes
        msg_attr = b""
        # attrib
        for cur_attr in m.attributes:
            attr_value = cur_attr.encode()
            data_attr =  struct.pack("!HH", cur_attr.attr_type, len(attr_value))
            data_attr += attr_value

            # padding ?
            while 4-(len(data_attr)% 4) != 4:
                data_attr += b"\x00"*(4-(len(data_attr)% 4))
            msg_attr += data_attr

        attr_length = len(msg_attr)

        # add message class and method
        stuntype = (((m.msgclass & 0x02) << 7) | ((m.msgclass & 0x01) << 4)) | m.msgmethod & 0x3EEF
        buf = struct.pack("!H", stuntype)

        # append message size
        buf += struct.pack("!H", attr_length)
        if m.magic_cookie > 0:
            buf += struct.pack("!L", m.magic_cookie)
            buf += struct.pack("!12s", m.transaction_id)
        else:
            buf += struct.pack("!16s", m.transaction_id)

        # append attributes
        buf += msg_attr

        return buf

    def send(self, data):
        """send data"""
        pass


from aiostun.client import Client
from aiostun.nat import NAT
from aiostun.stun import Codec
from aiostun.stun import Message

from aiostun.attribute import AttrChangeRequest

from aiostun.constants import FAMILY_IP4 as IP4
from aiostun.constants import FAMILY_IP6 as IP6

from aiostun.constants import IPPROTO_UDP as UDP
from aiostun.constants import IPPROTO_TCP as TCP
from aiostun.constants import IPPROTO_TLS as TLS

from aiostun.constants import CLASS_REQUEST
from aiostun.constants import CLASS_INDICATION

from aiostun.constants import METHOD_BINDING
from aiostun.constants import METHOD_SHARED_SECRET
from aiostun.constants import METHOD_ALLOCATE
from aiostun.constants import METHOD_REFRESH



import ssl
import asyncio
import socket

from aiostun import constants
from aiostun import stun
from aiostun import attribute

class TransportProtocol:
    def __init__(self, client, proto):
        """init"""
        self._client = client
        self._transport = None
        self._proto = proto

    def connection_made(self, transport):
        """on connection made"""
        self._transport = transport
        self._client.send = self.send

    def data_received(self, data):
        """on tcp/tls data received"""
        self._client.feed_data(data=data)

    def datagram_received(self, data, addr):
        """on udp datagram received"""
        self._client.feed_data(data)

    def send(self, data, addr=None):
        """send"""
        if self._proto == constants.IPPROTO_UDP:
            self._transport.sendto(data, addr=addr)
        if self._proto in [ constants.IPPROTO_TCP, constants.IPPROTO_TLS ]:
            self._transport.write(data)

    def error_received(self, exc):
        """on error"""
        print('error:', exc)

    def connection_lost(self, exc):
        """on connection lost"""
        pass


class Client:
    def __init__(self, host, port=3478, family=constants.FAMILY_IP4, proto=constants.IPPROTO_UDP, timeout=2):
        """init"""
        self._host = host
        self._port = port
        self._family = family
        self._ipproto = proto
        self._stun_codec = stun.Codec()
        self._transport = None
        self._timeout = timeout

    async def __aenter__(self):
        """aenter"""
        return await self.connect()

    async def __aexit__(self, exc_type, exc, tb):
        """aexit"""
        self.close()

    async def connect(self, remote_addr=True):
        """connect to remote"""
        loop = asyncio.get_event_loop()
        kwargs = {}
        if self._family == constants.FAMILY_IP4:
            kwargs['family'] = socket.AF_INET
        if self._family == constants.FAMILY_IP6:
            kwargs['family'] = socket.AF_INET6

        if self._ipproto == constants.IPPROTO_UDP:
            if remote_addr: kwargs['remote_addr']= self._host, self._port
            protocol = TransportProtocol(self._stun_codec, self._ipproto)
            kwargs['protocol_factory'] = lambda: protocol
            coro = loop.create_datagram_endpoint(**kwargs)

        if self._ipproto == constants.IPPROTO_TCP:
            kwargs['host']= self._host
            kwargs['port'] = self._port
            protocol = TransportProtocol(self._stun_codec, self._ipproto)
            kwargs['protocol_factory'] = lambda: protocol
            coro = loop.create_connection(**kwargs)

        if self._ipproto == constants.IPPROTO_TLS:
            sslcontext = ssl.create_default_context()
            sslcontext.check_hostname = False
            sslcontext.verify_mode = ssl.CERT_NONE
            kwargs['host']= self._host
            kwargs['port'] = self._port
            kwargs['ssl'] = sslcontext
            protocol = TransportProtocol(self._stun_codec, self._ipproto)
            kwargs['protocol_factory'] = lambda: protocol
            coro = loop.create_connection(**kwargs)

        try:
            self._transport, _ = await asyncio.wait_for(coro, timeout=self._timeout)
        except asyncio.TimeoutError:
            return self
        return self

    def close(self):
        """close transport"""
        if self._transport is not None:
            self._transport.close()

    def get_local_addr(self):
        """get local ip and port"""
        if self._transport is None:
            return None

        sock = self._transport.get_extra_info('socket')
        if sock is None:
            return None

        return sock.getsockname()

    def get_remote_addr(self):
        """get remote ip and port"""
        if self._transport is None:
            return None

        sock = self._transport.get_extra_info('socket')
        if sock is None:
            return None

        return sock.getpeername()

    def send_request(self, req, remote_addr=None):
        """wait request"""
        if self._transport is None:
            return False

        # encode stun message
        data = self._stun_codec.encode(req)

        # send it
        self._stun_codec.send(data=data, addr=remote_addr)

        return True

    async def wait_for_resp(self):
        """wait for response"""
        if self._transport is None:
            return None

        try:
            resp = await asyncio.wait_for(self._stun_codec._queue.get(), timeout=self._timeout)
        except asyncio.TimeoutError:
            return None
        return resp

    async def bind_request(self, use_classicstun=False, attrs=[], remote_addr=None):
        """send bind request"""
        stun_proto = stun.Message
        if use_classicstun:
            stun_proto = stun.ClassicMessage

        # basic stun message
        stun_req = stun_proto(constants.CLASS_REQUEST, constants.METHOD_BINDING, attrs)

        # send it
        success = self.send_request(req=stun_req, remote_addr=remote_addr)
        if not success:
            return {}

        # wait for response
        resp = await self.wait_for_resp()
        if resp is None:
            return None

        #If the message class is "Success Response" or "Error Response"
        # checks that the transaction ID matches the request
        if resp.msgclass in [ constants.CLASS_SUCCESS, constants.CLASS_ERROR]:
            if stun_req.transaction_id != resp.transaction_id:
                return None

        return resp

    async def get_mapped_address(self, use_classicstun=False):
        """get mapped address"""
        mapped_addr = {}

        resp = await self.bind_request(use_classicstun=use_classicstun)
        if resp is None:
            return mapped_addr

        attr = resp.get_attribute(attribute.AttrXorMappedAddr)
        if attr is None:
            attr = resp.get_attribute(attribute.AttrMappedAddr)
            if attr is None:
                return mapped_addr

        return attr.params



#!/usr/bin/python

import setuptools

with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

KEYWORDS = ('stun client')

setuptools.setup(
    name="aiostun",
    version="0.4.1",
    author="Denis MACHARD",
    author_email="d.machard@gmail.com",
    description="Asynchronous STUN client for Python with UDP, TCP and TLS support",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/dmachard/python-aiostun",
    packages=['aiostun', 'tests'],
    include_package_data=True,
    platforms='any',
    keywords=KEYWORDS,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries",
    ],
    install_requires=[]
)